# Generated by Django 3.1.8 on 2021-05-06 17:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cms_automations', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='automationhookplugin',
            name='operation',
            field=models.IntegerField(choices=[(0, 'Automatisierung starten'), (1, 'Nachricht an Automatisierung')], default=1, verbose_name='Operation'),
        ),
        migrations.AlterField(
            model_name='automationhookplugin',
            name='automation',
            field=models.CharField(max_length=128, verbose_name='Automatisierung'),
        ),
        migrations.AlterField(
            model_name='automationhookplugin',
            name='token',
            field=models.CharField(blank=True, max_length=128, verbose_name='Optionales Token'),
        ),
        migrations.AlterField(
            model_name='automationstatusplugin',
            name='template',
            field=models.CharField(max_length=128, verbose_name='Daten der Aufgabe'),
        ),
    ]



# coding=utf-8

from django.conf import settings
from django.utils.translation import gettext_lazy as _

MAX_FIELD_LENGTH = 128

FORM_VIEW_CONTEXT = getattr(settings, 'ATM_FORM_VIEW_CONTEXT',
                            dict(
                                submit_classes="btn btn-primary float-right",
                                back_classes="btn btn-outline-primary",
                            ))

TASK_LIST_TEMPLATES = getattr(settings, 'ATM_TASK_LIST_TEMPLATES',
                              (
                                  ('automations/includes/task_list.html', _("Default template")),
                              ))



# coding=utf-8

# Create your views here.
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.views.generic import TemplateView, FormView
from django.forms import BaseForm

from . import models, flow, settings


class AutomationMixin:
    _automation_instance = None  # Buffer class for specific task_instance
    _task_id = None

    def get_automation_instance(self, task):
        if self._automation_instance is None or self._task_id != task.id:
            cls = task.automation.get_automation_class()
            self._automation_instance = cls(automation=task.automation)
            self._task_id = task.id
        return self._automation_instance


class TaskView(LoginRequiredMixin, AutomationMixin, FormView):
    def bind_to_node(self):
        self.task = get_object_or_404(models.AutomationTaskModel, id=self.kwargs["task_id"])
        self.node = self.task.get_node()

    def get_form_kwargs(self):
        if not hasattr(self, "node"):
            self.bind_to_node()
        kwargs = super().get_form_kwargs()
        task_kwargs = self.node._form_kwargs
        kwargs.update(task_kwargs(self.task) if callable(task_kwargs) else task_kwargs)
        return kwargs

    def get_form_class(self):
        if not hasattr(self, "node"):
            self.bind_to_node()
        form = self.node._form
        return form if issubclass(form, BaseForm) else form(self.task)

    def get_context_data(self, **kwargs):
        if not hasattr(self, "node"):
            self.bind_to_node()
        if not isinstance(self.node, flow.Form):
            raise Http404
        if self.request.user not in self.task.get_users_with_permission():
            raise PermissionDenied
        if self.task.finished:
            raise Http404  # Need to display a message: s.o. else has completed form
        self.template_name = (
                self.node._template_name or
                getattr(self.get_automation_instance(self.task), 'default_template_name',
                        'automations/form_view.html'))
        context = super().get_context_data(**kwargs)
        context.update(settings.FORM_VIEW_CONTEXT)
        context.update(getattr(self.node._automation, 'context', dict()))
        context.update(self.node._context)
        return context

    def form_valid(self, form):
        self.node.validate(self.task, self.request, form)
        if self.node._run:
            self.node._automation.run(self.task.previous, self.node)
        if hasattr(self.node, "_success_url"):
            return redirect(self.node._success_url)
        return super().form_valid(form)

    def get_success_url(self):
        return reverse("task_list")


class TaskListView(LoginRequiredMixin, TemplateView):
    template_name = 'automations/task_list.html'

    def get_context_data(self, **kwargs):
        qs = models.AutomationTaskModel.get_open_tasks(self.request.user)
        return dict(error="", tasks=qs, count=len(qs))


class TaskDashboardView(TemplateView):
    template_name = 'automations/dashboard.html'

    def get_context_data(self, **kwargs):
        return dict(automations=models.AutomationModel.objects.all(),
                    tasks=models.AutomationTaskModel.objects.all())



from cms.models import CMSPlugin
from django.utils.translation import gettext as _

from .. import settings
from django.db import models


class AutomationTasksPlugin(CMSPlugin):
    template = models.CharField(
        max_length=settings.MAX_FIELD_LENGTH,
        choices=settings.TASK_LIST_TEMPLATES,
        default=settings.TASK_LIST_TEMPLATES[0][0],
        blank=False,
        verbose_name=_("Template"),
    )
    always_inform=models.BooleanField(
        default=True,
        verbose_name=_("Always inform"),
        help_text=_("If deactivated plugin will out output anything if no task is available.")
    )


class AutomationHookPlugin(CMSPlugin):          # pragma: no cover
    class OperationChoices(models.IntegerChoices):
        start = 0, _("Start automaton")
        message = 1, _("Send message to automation")

    operation = models.IntegerField(
        default=OperationChoices.message,
        choices=OperationChoices.choices,
        verbose_name=_("Operation")
    )

    automation = models.CharField(
        blank=False,
        max_length=settings.MAX_FIELD_LENGTH,
        verbose_name=_("Automation"),
    )

    token = models.CharField(
        max_length=settings.MAX_FIELD_LENGTH,
        blank=True,
        verbose_name=_("Optional token"),
    )

    def __str__(self):
        return self.automation.split('.')[-1]

class AutomationStatusPlugin(CMSPlugin):            # pragma: no cover
    template = models.CharField(
        blank=False,
        max_length=settings.MAX_FIELD_LENGTH,
        verbose_name=_("Task data"),
    )
    name = models.CharField(
        blank=True,
        max_length=settings.MAX_FIELD_LENGTH,
    )

    def __str__(self):
        return self.name if self.name else super().__str__()



# coding=utf-8

"""Soft dependecy on django-automations_cms: Define plugins for open tasks"""
import logging

from django import forms

from .. import models
from . import models as cms_models
from .. import flow


try:
    from cms.plugin_base import CMSPluginBase
    from cms.utils import get_current_site
    from cms.plugin_pool import plugin_pool
    from django.utils.translation import gettext_lazy as _

    logger = logging.getLogger(__name__)

    class AutomationTaskList(CMSPluginBase):
        name = _("Task list")
        module = _("Automations")
        model = cms_models.AutomationTasksPlugin
        allow_children = False
        require_parent = False
        render_template = None

        def render(self, context, instance, placeholder):
            self.render_template = instance.template
            qs = models.AutomationTaskModel.get_open_tasks(context['request'].user)
            context.update(dict(tasks=qs, count=len(qs), always_inform=instance.always_inform))
            return context

    plugin_pool.register_plugin(AutomationTaskList)


    def get_task_choices(pattern, convert, subcls=None):
        status_choices = []
        for cls_name, verbose_name in flow.get_automations():
            cls = models.get_automation_class(cls_name)
            choices = []
            if subcls is not None and hasattr(cls, subcls):
                cls = getattr(cls, subcls)
            for item in dir(cls):
                if pattern(item):
                    attr = getattr(cls, item)
                    tpl = convert(attr, item, cls_name)
                    if isinstance(tpl, (tuple, list)):
                        choices.append(tuple(tpl))
            if choices:
                status_choices.append((verbose_name, tuple(choices)))
        return tuple(status_choices)   # make immutable

    def get_task_status_choices():
        def convert(attr, item, cls_name):
            if isinstance(attr, str):
                attr = attr, item.replace('_', ' ').capitalize()
            return attr
        return get_task_choices(lambda x: x.endswith("_template"), convert=convert, subcls="Meta")

    def get_task_receiver_choices():
        def convert(attr, item, cls_name):
            if callable(attr) and len(item) > 8:
                return cls_name+'.'+item[8:], item[8:].replace('_', ' ').capitalize()
            return None
        return get_task_choices(lambda x: x.startswith("receive_"), convert=convert)

    def get_int(get_data, data):
        if data in get_data and isinstance(get_data[data], str) and get_data[data].isnumeric():
            return int(get_data[data])
        return None

    def get_automation_instance(get_params):
        task_id, atm_id = get_int(get_params, "task_id"), get_int(get_params, "atm_id")
        if atm_id is not None:
            try:
                automation_instance = models.AutomationModel.objects.get(id=atm_id)
            except models.AutomationModel.DoesNotExist:
                return None
        elif task_id is not None:
            try:
                automation_instance = models.AutomationTaskModel.objects.get(id=atm_id).automation
            except models.AutomationTaskModel.DoesNotExist:
                return None
        else:
            return None

    def get_automation_data(context, template):
        automation_instance = get_automation_instance(context.get("request", dict()).GET)
        if automation_instance:
            cls = models.get_automation_class(automation_instance.automation_class)
            if hasattr(cls, "Meta"):
                for item in dir(cls.Meta):
                    if item.endswith('_template'):
                        attr = getattr(cls.Meta, item)
                        if (isinstance(attr, str) and attr == template or
                                isinstance(attr, (tuple, list)) and attr[0] == template):
                            return automation_instance
        return None


    class EditTaskData(forms.ModelForm):
        class Meta:
            model = cms_models.AutomationStatusPlugin
            widgets = {
                'template': forms.Select(choices=get_task_status_choices()),
                'name': forms.HiddenInput(),
            }
            fields = "__all__"

        def clean_name(self):
            choices = {}
            for _, chapter in get_task_status_choices():
                choices.update({key: value for key, value in chapter})
            return choices.get(self.data['template'], "")


    class AutomationStatus(CMSPluginBase):
        name = _("Status")
        module = _("Automations")
        model = cms_models.AutomationStatusPlugin
        allow_children = False
        require_parent = False
        text_enabled = True
        form = EditTaskData
        render_template = None

        def render(self, context, instance, placeholder):
            self.render_template = instance.template

            automation_instance = get_automation_data(context, self.render_template)
            if automation_instance is not None:
                automation = automation_instance.get_automation_class()(automation_id=automation_instance.id)
            else:
                automation = None
            context.update(dict(automation=automation,
                                automation_instance=automation_instance,
                                instance=instance))
            return context

    plugin_pool.register_plugin(AutomationStatus)


    class EditAutomationHook(forms.ModelForm):
        class Meta:
            model = cms_models.AutomationHookPlugin
            widgets = {
                'automation':   forms.Select(choices=get_task_receiver_choices()),
            }
            fields = "__all__"


    class AutomationHook(CMSPluginBase):
        name = _("Send message")
        module = _("Automations")
        model = cms_models.AutomationHookPlugin
        allow_children = False
        require_parent = False
        render_template = "automations/cms/empty_template.html"
        form = EditAutomationHook

        def render(self, context, instance, placeholder):
            request = context['request']
            automation, message = instance.automation.rsplit('.', 1)
            cls = models.get_automation_class(automation)
            if instance.operation == cms_models.AutomationHookPlugin.OperationChoices.message:
                automation_id = request.GET.get("atm_id", None)
                if isinstance(automation_id, str) and automation_id.isnumeric():
                    cls.dispatch_message(int(automation_id), message, instance.token, request)
                else:
                    logger.error("Invalid automation instance: %s" % instance)
            else:
                cls.create_on_message(message, instance.token, request)
            return context


    plugin_pool.register_plugin(AutomationHook)


except ImportError:
    pass



# coding=utf-8
from io import StringIO
from unittest.mock import patch

from django import forms
from django.contrib.auth.models import User
from django.core.management import execute_from_command_line
from django.test import TestCase
import datetime

from django.utils.timezone import now
from django.utils.translation import gettext as _

from . import flow, models, views
from .flow import this

import django.dispatch
from django.test import RequestFactory


# Create your tests here.

test_signal = django.dispatch.Signal()


class Print(flow.Execute):
    @staticmethod
    def method(task_instance, *args):
        print(task_instance.status, *args)


class TestAutomation(flow.Automation):
    start =             flow.Execute(this.init).AsSoonAs(lambda x:True).AsSoonAs(this.cont)
    intermediate =      flow.Execute("self.init2")
    func_if =           flow.If(lambda x: x.data['more_participants'] == "test").Then().Else(this.print)
    if_clause =         flow.If(lambda x: x.data['more_participants'] == "test").Then("self.conditional")
    if2 =               flow.If(lambda x: x.data['more_participants'] == "test"
                                ).Then(this.if_clause
                                ).Else(this.end)
    end =               flow.End()

    conditional =       flow.Execute(this.init2).Next("self.if2")

    def init(self, task_instance, *args, **kwargs):
        if 'participants' not in self.data:
            self.data['participants'] = []
            self.save()

    def init2(self, task):
        self.data['more_participants'] = "test" + self.data.get("more_participants", "")
        self.save()
        return task  # Illegal since not json serializable

    def cont(self, task):
        return bool(self) and (task)  # True

    def print(self, task):
        print("Hello", task.data)


class TestForm(forms.Form):
    success_url = "https://www.google.de/"
    first_name = forms.CharField(
            label=_('First name'),
            max_length=80,
    )
    email = forms.EmailField(
        label=_('Your e-mail address'),
        max_length=100,
    )
    session = forms.IntegerField(
            label=_('Chose session'),
    )


class TestSplitJoin(flow.Automation):
    class Meta:
        verbose_name = "Allow to split and join"
        verbose_name_plural = "Allow splitS and joinS"

    start = Print("Hello, this is the single thread").AfterPausingFor(datetime.timedelta(days=-1))
    l10 = Print("Line 10").AfterWaitingUntil(now()-datetime.timedelta(minutes=1))
    split = flow.Split().Next("self.t10").Next("self.t20").Next("self.t30")
    join = flow.Join()
    l20 = Print("All joined now")
    l30 = flow.End()

    t10 = Print("Thread 10").Next(this.join)
    t20 = Print("Thread 20").Next(this.join)
    t30 = Print("Thread 30").Next(this.join)


@flow.on_signal(test_signal)
class SignalAutomation(flow.Automation):
    def started_by_signal(self, *args, **kwargs):
        self.data['started'] = 'yeah!'
        self.save()
    start = flow.Execute().AfterPausingFor(datetime.timedelta(days=1))
    end = flow.End()

    def receive_new_user(self, token, data=None):
        self.data['token'] = token
        self.save()
        return "received"


class SendMessageAutomation(flow.Automation):
    start = flow.SendMessage(SignalAutomation, "new_user", "12345678")
    to_nowhere = flow.SendMessage("automations.tests.FormTest", "this_receiver_does_not_exist")
    end = flow.End()


class AtmTaskForm(forms.ModelForm):
    class Meta:
        model = models.AutomationTaskModel
        exclude = []


class FormTest(flow.Automation):
    class Meta:
        verbose_name = "Edit task #8"
        verbose_name_plural = "Edit tasks"

    start = flow.Execute(this.init_with_item)
    form = (flow.Form(TestForm, context=dict(claim="Save"))
            .User(id=0))
    form2 = (flow.Form(TestForm, context=dict(claim="Save")).User(id=0)
            .Permission("automations.create_automationmodel"))
    end = flow.End()

    def init_with_item(self, task_instance):
        task_instance.data['task_id'] = 8
        self.save()


class Looping(flow.Automation):
    start = flow.Split().Next("self.loop1").Next("self.loop2").Next("self.loop3")

    loop1 = flow.ModelForm(AtmTaskForm, "key_id")
    loop1_1 = flow.Repeat("self.loop1").EveryDay().At(21,00)
    loop2 = flow.Repeat("self.loop2").EveryNMinutes(30)
    loop3 = flow.Repeat("self.loop3").EveryHour()


class BoundToFail(flow.Automation):
    start = Print("Will divide by zero.")
    div = flow.Execute(lambda x: 5/0).OnError(this.error_node)
    never = Print("This should NOT be printed")
    not_caught = flow.Execute(lambda x: 5/0)
    never_reached = Print("Will not reach this")
    end = flow.End()

    error_node = Print("Oh dear").Next(this.not_caught)


class SingletonAutomation(flow.Automation):
    singleton = True

    end = flow.End()


class ByEmailSingletonAutomation(flow.Automation):
    singleton = ("email", )

    end = flow.End()

    @flow.require_get_parameters(email=str, mails=int)
    def receive_test(self, token, data):
        pass

class ModelTestCase(TestCase):

    def test_modelsetup(self):
        x = TestAutomation(autorun=False)
        qs = models.AutomationModel.objects.all()
        self.assertEqual(len(qs), 1)
        self.assertEqual(qs[0].automation_class, 'automations.tests.TestAutomation')

        self.assertEqual(models.get_automation_class(x._db.automation_class), TestAutomation)

        x.run()
        self.assertIn("more_participants", x.data)
        self.assertEqual(x.data["more_participants"], "testtest")
        self.assertIn("participants", x.data)

        self.assertEqual(x.get_verbose_name(), "Automation TestAutomation")
        self.assertEqual(x.get_verbose_name_plural(), "Automations TestAutomation")



class AutomationTestCase(TestCase):

    def test_split_join(self):
        with patch('sys.stdout', new=StringIO()) as fake_out:
            atm = TestSplitJoin()
        output = fake_out.getvalue().splitlines()
        self.assertEqual(output[0], "start Hello, this is the single thread")
        self.assertEqual(output[1], "l10 Line 10")
        self.assertEqual(output[-1], "l20 All joined now")
        self.assertIn("t10 Thread 10", output)
        self.assertIn("t20 Thread 20", output)
        self.assertIn("t30 Thread 30", output)

        self.assertEqual(atm.get_verbose_name(), "Allow to split and join")
        self.assertEqual(atm.get_verbose_name_plural(), "Allow splitS and joinS")


class FormTestCase(TestCase):
    def setUp(self):
        # Every test needs access to the request factory.
        self.factory = RequestFactory()
        self.user = User.objects.create_user(
            username='jacob', email='jacob@…', password='top_secret')

    def test_form(self):
        atm = FormTest(autorun=False)
        atm.form._user=dict(id=self.user.id)  # Fake User
        atm.form2._user=dict(id=self.user.id)  # Fake User
        atm.run()

        users = atm.form.get_users_with_permission()
        self.assertEqual(len(users), 0)

        tasks = atm._db.automationtaskmodel_set.filter(finished=None)
        self.assertEqual(len(tasks), 1)

        # Create an instance of a GET request.
        request = self.factory.get(f"/{tasks[0].id}")

        # Recall that middleware are not supported. You can simulate a
        # logged-in user by setting request.user manually.
        request.user = self.user
        response = views.TaskView.as_view()(request, task_id=tasks[0].id)
        self.assertEqual(response.status_code, 200)

        form_data = dict(session=8, email="none@nowhere.com", first_name="Fred")
        request = self.factory.post(f"/{tasks[0].id}", data=form_data)
        request.user = self.user
        response = views.TaskView.as_view()(request, task_id=tasks[0].id)
        self.assertEqual(response.status_code, 302)  # Success leads to redirect

        request = self.factory.get(f"/tasks")
        request.user = self.user
        response = views.TaskListView.as_view()(request)
        self.assertEqual(response.status_code, 200)


        request = self.factory.get(f"/tasks")
        request.user = self.user
        response = views.TaskListView.as_view()(request)
        self.assertEqual(response.status_code, 200)

        atm.run()
        self.assertEqual(len(atm.form2.get_users_with_permission()), 0)


class SignalTestCase(TestCase):
    def test_signal(self):
        self.assertEqual(0, len(models.AutomationModel.objects.filter(
            automation_class='automations.tests.SignalAutomation',
        )))

        test_signal.send(self.__class__)

        inst = models.AutomationModel.objects.filter(
            automation_class='automations.tests.SignalAutomation',
        )
        self.assertEqual(1, len(inst))
        self.assertEqual(inst[0].data.get("started", ""), "yeah!")
        sender = SendMessageAutomation()
        inst = models.AutomationModel.objects.filter(
            automation_class='automations.tests.SignalAutomation',
        )
        self.assertEqual(inst[0].data.get("token", ""), "12345678")

        self.assertGreater(len(inst[0].automationtaskmodel_set.all()), 0)


class RepeatTest(TestCase):
    def test_repeat(self):
        atm = Looping()
        tasks = atm._db.automationtaskmodel_set.filter(finished=None)
        self.assertEqual(len(tasks), 3)
        atm.run()
        tasks = atm._db.automationtaskmodel_set.filter(finished=None)
        self.assertEqual(len(tasks), 3)

    def test_get_automations(self):
        self.assertEqual(len(flow.get_automations()), 0)
        self.assertEqual(len(flow.get_automations("automations.flow")), 1)
        tpl = flow.get_automations("automations.tests")
        self.assertIn('Allow to split and join', (name for _, name in tpl))


class ManagementCommandTest(TestCase):
    def test_managment_command(self):
        with patch('sys.stdout', new=StringIO()) as fake_out:
            atm = TestSplitJoin()
            execute_from_command_line(["manage.py", "automation_step"])
        output = fake_out.getvalue().splitlines()
        self.assertEqual(output[0], "start Hello, this is the single thread")
        self.assertEqual(output[1], "l10 Line 10")
        self.assertEqual(output[-1], "l20 All joined now")
        self.assertIn("t10 Thread 10", output)
        self.assertIn("t20 Thread 20", output)
        self.assertIn("t30 Thread 30", output)

        db_id = atm._db.id
        self.assertGreater(len(models.AutomationTaskModel.objects.filter(
            automation_id=db_id
        )), 0)
        atm.kill()
        self.assertEqual(len(models.AutomationTaskModel.objects.filter(
            automation_id=db_id
        )), 0)


class ExecutionErrorTest(TestCase):
    def test_managment_command(self):
        with patch('sys.stdout', new=StringIO()) as fake_out:
            atm = BoundToFail()
        output = fake_out.getvalue().splitlines()
        self.assertEqual(output[0], "start Will divide by zero.")
        self.assertEqual(output[-1], "error_node Oh dear")
        self.assertNotIn("never This should NOT be printed", output)
        self.assertNotIn("never_reached Will not reach this", output)


class SingletonTest(TestCase):
    def test_singleton(self):
        inst1 = SingletonAutomation()
        self.assertEqual(len(models.AutomationModel.objects.filter(
            automation_class="automations.tests.SingletonAutomation"
        )), 1)
        inst2 = SingletonAutomation()
        self.assertEqual(len(models.AutomationModel.objects.filter(
            automation_class="automations.tests.SingletonAutomation"
        )), 1)
        self.assertEqual(inst1._db, inst2._db)
        self.assertNotEqual(inst1, inst2)

    def test_rel_singleton(self):
        inst1 = ByEmailSingletonAutomation(email="none@nowhere.com",autorun=False)
        atm_name = inst1.get_automation_class_name()
        self.assertEqual(atm_name.rsplit(".", 1)[-1], inst1.end.get_automation_name())
        self.assertEqual(len(models.AutomationModel.objects.filter(
            automation_class=atm_name
        )), 1)
        inst2 = ByEmailSingletonAutomation(email="nowhere@none.com",autorun=False)
        self.assertEqual(len(models.AutomationModel.objects.filter(
            automation_class=atm_name
        )), 2)
        self.assertNotEqual(inst1._db, inst2._db)
        inst3 = ByEmailSingletonAutomation(email="nowhere@none.com", autorun=False)
        self.assertEqual(len(models.AutomationModel.objects.filter(
            automation_class=atm_name
        )), 2)
        self.assertEqual(inst2._db, inst3._db)

        self.assertTrue(ByEmailSingletonAutomation.satisfies_data_requirements("test", dict(email="test", mails="2")))
        self.assertTrue(ByEmailSingletonAutomation.satisfies_data_requirements("test", dict(email="test", mails=2)))
        self.assertFalse(ByEmailSingletonAutomation.satisfies_data_requirements("test", dict(email="test", mails="t2")))
        self.assertFalse(ByEmailSingletonAutomation.satisfies_data_requirements("test", dict(mails="t2")))
        self.assertFalse(ByEmailSingletonAutomation.satisfies_data_requirements("nonexistent", dict(mails="t2")))

        ByEmailSingletonAutomation.create_on_message("test", None, dict(email="new", mails=2))
        self.assertEqual(len(models.AutomationModel.objects.filter(
            automation_class="automations.tests.ByEmailSingletonAutomation"
        )), 3)

        ByEmailSingletonAutomation.create_on_message("test", None, dict(email="also_new"))
        self.assertEqual(len(models.AutomationModel.objects.filter(
            automation_class="automations.tests.ByEmailSingletonAutomation"
        )), 3)

        models.AutomationModel.run()





# Generated by Django 3.1.8 on 2021-05-02 08:56

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('auth', '0012_alter_user_first_name_max_length'),
    ]

    operations = [
        migrations.CreateModel(
            name='AutomationModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('automation_class', models.CharField(max_length=256, verbose_name='Process class')),
                ('finished', models.BooleanField(default=False, verbose_name='Finished')),
                ('data', models.JSONField(default=dict, verbose_name='Data')),
                ('paused_until', models.DateTimeField(null=True, verbose_name='Paused until')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name='AutomationTaskModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.CharField(blank=True, max_length=256, verbose_name='Status')),
                ('locked', models.IntegerField(default=0, verbose_name='Locked')),
                ('interaction_permissions', models.JSONField(default=list, help_text='List of permissions of the form app_label.codename', verbose_name='Required permissions')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('finished', models.DateTimeField(null=True)),
                ('message', models.CharField(blank=True, max_length=128, verbose_name='Message')),
                ('result', models.JSONField(blank=True, default=dict, null=True, verbose_name='Result')),
                ('automation', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='automations.automationmodel')),
                ('interaction_group', models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='auth.group', verbose_name='Assigned group')),
                ('interaction_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL, verbose_name='Assigned user')),
                ('previous', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='automations.automationtaskmodel', verbose_name='Previous task')),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
from django import template
from .. import models

register = template.Library()



# @register.simple_tag(takes_context=True)
# def task_info(context, automation_class, info, debug=False):
#     get_data = context.get("request", dict()).GET
#     task_id, atm_id = get_int(get_data, "task_id"), get_int(get_data, "atm_id")
#     if automation is not None and automation.automation_class == automation_class:
#         cls = automation.get_automation_class()
#         if hasattr(cls, "public_data") and info in cls.public_data:
#             return automation.data.get(info, "")
#     return "-- Task info --" if debug else ""



# Generated by Django 3.1.8 on 2021-05-02 11:09

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('cms', '0022_auto_20180620_1551'),
    ]

    operations = [
        migrations.CreateModel(
            name='AutomationHookPlugin',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='cms_automations_automationhookplugin', serialize=False, to='cms.cmsplugin')),
                ('automation', models.CharField(max_length=128, verbose_name='Automation')),
                ('token', models.CharField(blank=True, max_length=128, verbose_name='Optional token')),
            ],
            options={
                'abstract': False,
            },
            bases=('cms.cmsplugin',),
        ),
        migrations.CreateModel(
            name='AutomationStatusPlugin',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='cms_automations_automationstatusplugin', serialize=False, to='cms.cmsplugin')),
                ('template', models.CharField(max_length=128, verbose_name='Task data')),
                ('name', models.CharField(blank=True, max_length=128)),
            ],
            options={
                'abstract': False,
            },
            bases=('cms.cmsplugin',),
        ),
        migrations.CreateModel(
            name='AutomationTasksPlugin',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='cms_automations_automationtasksplugin', serialize=False, to='cms.cmsplugin')),
                ('template', models.CharField(choices=[('automations/includes/task_list.html', 'Default template')], default='automations/includes/task_list.html', max_length=128, verbose_name='Template')),
                ('always_inform', models.BooleanField(default=True, help_text='If deactivated plugin will out output anything if no task is available.', verbose_name='Always inform')),
            ],
            options={
                'abstract': False,
            },
            bases=('cms.cmsplugin',),
        ),
    ]



# coding=utf-8
from django.urls import path

from . import views

urlpatterns = [
    path("<int:task_id>", views.TaskView.as_view(), name="task"),
    path("tasks", views.TaskListView.as_view(), name="task_list"),
    path("dashboard", views.TaskDashboardView.as_view(), name="dashboard"),
]



# Generated by Django 3.1.8 on 2021-05-06 17:57

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('automations', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='automationmodel',
            name='automation_class',
            field=models.CharField(max_length=256, verbose_name='Prezess-Klasse'),
        ),
        migrations.AlterField(
            model_name='automationmodel',
            name='data',
            field=models.JSONField(default=dict, verbose_name='Daten'),
        ),
        migrations.AlterField(
            model_name='automationmodel',
            name='finished',
            field=models.BooleanField(default=False, verbose_name='Beendet'),
        ),
        migrations.AlterField(
            model_name='automationmodel',
            name='paused_until',
            field=models.DateTimeField(null=True, verbose_name='Pausiert bis'),
        ),
        migrations.AlterField(
            model_name='automationtaskmodel',
            name='locked',
            field=models.IntegerField(default=0, verbose_name='Gesperrt'),
        ),
        migrations.AlterField(
            model_name='automationtaskmodel',
            name='message',
            field=models.CharField(blank=True, max_length=128, verbose_name='Nachricht'),
        ),
        migrations.AlterField(
            model_name='automationtaskmodel',
            name='previous',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='automations.automationtaskmodel', verbose_name='Vorherige Aufgabe'),
        ),
        migrations.AlterField(
            model_name='automationtaskmodel',
            name='result',
            field=models.JSONField(blank=True, default=dict, null=True, verbose_name='Ergebnis'),
        ),
    ]



# coding=utf-8
import datetime
import functools
import json
import sys
import threading
import traceback
from copy import copy

from django.contrib.auth.models import User, Group, Permission
from django.core.exceptions import ImproperlyConfigured
from django.db.models import Q
from django.db.models.base import ModelBase
from django.db.transaction import atomic
from django.utils.timezone import now

from . import models, settings

"""To allow forward references in Automation object "this" is defined"""


class ThisAttribute:
    """Wrapper for forward-reference to a named attribute"""

    def __init__(self, attr):
        self.attr = attr

    def __repr__(self):
        return f"this.{self.attr}"


class This:
    """Generator for reference to a named attribute"""

    def __getattr__(self, item):
        return ThisAttribute(item)


this = This()
"""Global instance"""

"""
"""


def on_execution_path(m):
    """Decorator to ensure automatic pausing of automations in
    case of WaitUntil, PauseFor and When"""
    @functools.wraps(m)
    def wrapper(self, task, *args, **kwargs):
        return None if task is None else m(self, task, *args, **kwargs)

    return wrapper


class Node:
    """Parent class for all nodes"""

    def __init__(self, *args, **kwargs):
        self._conditions = []
        self._next = None
        self._wait = None
        self.description = kwargs.pop('description', '')

    @staticmethod
    def eval(sth, task):
        return sth(task) if callable(sth) else sth

    def ready(self, automation_instance, name):
        """is called by the newly initialized Automation instance to bind the nodes to the instance."""
        self._automation = automation_instance
        self._name = name
        self._conditions = [self.resolve(condition) for condition in self._conditions]

    def get_automation_name(self):
        """returns the name of the Automation instance class the node is bound to"""
        return self._automation.__class__.__name__

    def __getattribute__(self, item):
        value = super().__getattribute__(item)
        if isinstance(value, ThisAttribute) or (isinstance(value, str) and value.startswith('self.')):
            value = self.resolve(value)
            setattr(self, item, value)  # remember
        return value

    def resolve(self, value):
        if isinstance(value, ThisAttribute):  # This object?
            value = getattr(self._automation, value.attr)  # get automation attribute
        elif isinstance(value, str) and value.startswith('self.'):  # String literal instead of this
            value = getattr(self._automation, value[5:])
        return value

    @atomic
    def enter(self, prev_task=None):
        assert prev_task is None or prev_task.finished is not None, "Node entered w/o previous node left"
        db = self._automation._db
        assert isinstance(db, models.AutomationModel)
        task, _ = db.automationtaskmodel_set.get_or_create(
            previous=prev_task,
            status=self._name,
            defaults=dict(
                locked=0,
            ),
        )
        if task.locked > 0:
            return None
        task.locked += 1
        task.save()
        return task

    @atomic
    def release_lock(self, task: models.AutomationTaskModel):
        task.locked -= 1
        task.save()
        return None

    @staticmethod
    def store_result(task: models.AutomationTaskModel, message, result):
        task.message = message[0:settings.MAX_FIELD_LENGTH]
        try:  # Check if result is json serializable
            json.dumps(result)  # Raise error if not json-serializable
            task.result = result  # if it is, store it
        except TypeError:
            task.result = None
        task.save()

    def leave(self, task: models.AutomationTaskModel):
        if task is not None:
            task.finished = now()
            self.release_lock(task)
            if self._next is None:
                next_node = self._automation._iter[self]
            else:
                next_node = self._next
            if next_node is None:
                raise ImproperlyConfigured("No End statement")
            return next_node

    @on_execution_path
    def when_handler(self, task):
        for condition in self._conditions:
            if not self.eval(condition, task):
                return self.release_lock(task)
        return task

    @on_execution_path
    def wait_handler(self, task: models.AutomationTaskModel):
        if self._wait is None:
            return task
        earliest_execution = self.eval(self._wait, task)
        if earliest_execution < now():
            return task
        if self._automation._db.paused_until:
            self._automation._db.paused_until = min(self._automation._db.paused_until, earliest_execution)
        else:
            self._automation._db.paused_until = earliest_execution
        self._automation._db.save()
        return self.release_lock(task)

    def execute(self, task: models.AutomationTaskModel):
        return self.when_handler(self.wait_handler(task))

    def Next(self, next_node):
        if self._next is not None:
            raise ImproperlyConfigured(f"Multiple .Next statements")
        self._next = next_node
        return self

    def AsSoonAs(self, condition):
        self._conditions.append(condition)
        return self

    def AfterWaitingUntil(self, time):
        if self._wait is not None:
            raise ImproperlyConfigured(f"Multiple .WaitUntil or .PauseFor statements")
        self._wait = time
        return self

    def AfterPausingFor(self, timedelta):
        if self._wait is not None:
            raise ImproperlyConfigured(f"Multiple .WaitUntil or .PauseFor statements")
        self._wait = lambda x: x.created + self.eval(timedelta, x)
        return self

    def __repr__(self):
        return f"<{f'{self._name}: ' if self._automation else ''}" \
               f"{self._automation if self._automation else 'unbound'} {self.__class__.__name__} node>"


class End(Node):
    def execute(self, task):
        self._automation._db.finished = True
        self._automation._db.save()
        return task

    def leave(self, task):
        task.finished = now()
        task.locked = 0;
        task.save()


class Repeat(Node):
    def __init__(self, start=None, **kwargs):
        if start is None:
            start = self._automation.start

        super().__init__(**kwargs)
        self._next = start
        self._interval = None
        self._startpoint = None

    @on_execution_path
    def repeat_handler(self, task):
        if self._startpoint is None:
            self._startpoint = now()
        elif now() < self._startpoint:
            return self.release_lock(task)
        db = self._automation._db
        if db.paused_until:
            if now() < db.paused_until:
                return self.release_lock(task)
        else:
            db.paused_until = self._startpoint
        while self._automation._db.paused_until < now():
            db.paused_until += self._interval
        db.save()
        return task

    def execute(self, task: models.AutomationTaskModel):
        task = super().execute(task)
        return self.repeat_handler(task)

    def At(self, hour, minute):
        if self._interval is None:
            raise ImproperlyConfigured(f"Repeat().At: interval statement necessary before")
        if self._interval < datetime.timedelta(days=1):
            raise ImproperlyConfigured(f"Repeat().At: interval >= one day required")
        if self._startpoint is not None:
            raise ImproperlyConfigured(f"Repeat(): Only one .At modifier possible")
        self._startpoint = now()
        self._startpoint.replace(hour=hour, minute=minute)
        return self

    def EveryHour(self, hours=1):
        if self._interval is not None:
            raise ImproperlyConfigured(f"Repeat(): Multiple interval statements")
        self._interval = datetime.timedelta(hours=hours)
        return self

    def EveryNMinutes(self, minutes):
        if self._interval is not None:
            raise ImproperlyConfigured(f"Repeat(): Multiple interval statements")
        self._interval = datetime.timedelta(minutes=minutes)
        return self

    def EveryNDays(self, days):
        if self._interval is not None:
            raise ImproperlyConfigured(f"Repeat(): Multiple interval statements")
        self._interval = datetime.timedelta(days=days)
        return self

    def EveryDay(self):
        return self.EveryNDays(days=1)


class Split(Node):
    """Spawn several tasks which have to be joined by a Join() node"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._splits = []

    def Next(self, node):
        self._splits.append(node)
        return self

    def execute(self, task: models.AutomationTaskModel):
        task = super().execute(task)
        if task:
            assert len(self._splits) > 0, "at least one .Next statement needed for Split()"
            db = self._automation._db
            tasks = list(
                db.automationtaskmodel_set.create(  # Create splits
                    previous=task,
                    status=self.resolve(split)._name,
                    locked=0,
                ) for split in self._splits)
            self.leave(task)
            for task in tasks:
                self._automation.run(task.previous, getattr(self._automation, task.status))  # Run other splits
            return None
        return task


class Join(Node):
    """Collect tasks spawned by Split"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def execute(self, task: models.AutomationTaskModel):
        task = super().execute(task)
        if task:
            split_task = self.get_split(task)
            if split_task is None:
                raise ImproperlyConfigured("Join() without Split()")
            all_splits = []
            for open_task in self._automation._db.automationtaskmodel_set.filter(
                    finished=None,
            ):
                split = self.get_split(open_task)
                if split and split.id == split_task.id:
                    all_splits.append(open_task)
            assert len(all_splits) > 0, "Internal error: at least one split expected"
            if len(all_splits) > 1:  # more than one split at the moment: close this split
                self.leave(task)
                return None
        return task

    def get_split(self, task):
        split_task = task.previous
        while split_task is not None:
            node = getattr(self._automation, split_task.status)
            if isinstance(node, Split):
                return split_task
            split_task = split_task.previous  # Go back the history
        return None


class Execute(Node):
    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        self._on_error = None
        self.args = args
        self.kwargs = kwargs
        self._err = None

    def method(self, task, *args, **kwargs):
        func = args[0]
        if not callable(func):
            raise ImproperlyConfigured(f"Execute: expected callable, got {func.__class__.__name__}")
        return func(task, *self.args[1:], **self.kwargs)

    @on_execution_path
    def execute_handler(self, task: models.AutomationTaskModel):
        def func(task, *args, **kwargs):
            self._err = None
            try:
                result = self.method(task, *args, **kwargs)
                self.store_result(task, "OK", result)
            except Exception as err:
                if isinstance(err, ImproperlyConfigured):
                    raise err
                self._err = err
                self.store_result(task, repr(err), dict(error=traceback.format_exc()))


        if self.args is not None and len(self.args) > 0:  # Empty arguments: No-op
            args = (self.resolve(value) for value in self.args)
            kwargs = {key: self.resolve(value) for key, value in self.kwargs.items()}
            if kwargs.get("threaded", False):
                assert self._on_error is None, "No .OnError statement on threaded executions"
                threading.Thread(target=func, args=[task] + list(args), kwargs=kwargs).start()
            else:
                func(task, *args, **kwargs)
                if self._err:
                    if self._on_error:
                        self._next = self._on_error
                    else:
                        self.leave(task)
                        self._automation._db.finished = True
                        self._automation._db.save()
                        return None
        return task

    def execute(self, task: models.AutomationTaskModel):
        task = super().execute(task)
        return self.execute_handler(task)

    def OnError(self, next_node):
        if self._on_error is not None:
            raise ImproperlyConfigured(f"Multiple .OnError statements")
        self._on_error = next_node
        return self


class If(Execute):
    def __init__(self, condition, **kwargs):
        super().__init__(None, **kwargs)
        self._condition = condition
        self._then = None
        self._else = None
        self._func = lambda x: None

    def Then(self, *clause_args, **clause_kwargs):
        if self._then is not None:
            raise ImproperlyConfigured(f"Multiple .Then statements")
        self._then = (clause_args, clause_kwargs)
        return self

    def Else(self, *clause_args, **clause_kwargs):
        if self._else is not None:
            raise ImproperlyConfigured(f"Multiple .Else statements")
        self._else = (clause_args, clause_kwargs)
        return self

    @on_execution_path
    def if_handler(self, task: models.AutomationTaskModel):
        if self._then is None:
            raise ImproperlyConfigured(f"Missing .Then statement")
        this_path = self.eval(self._condition, task)
        clause = self._then if this_path else self._else
        if clause is not None:
            opt_args, opt_kwargs = clause
            if len(opt_args) == 1 and len(opt_kwargs) == 0:
                resolved = self.resolve(opt_args[0])
                if isinstance(resolved, Node) and not callable(resolved):
                    self.Next(resolved)
                    return task
            self.args = opt_args
            self.kwargs = opt_kwargs
            return self.execute_handler(task)
        return task

    def execute(self, task: models.AutomationTaskModel):
        # Do not execute super() since If inherits from Execute and there
        # is nothing to execute, call Node.execute instead
        task = Node.execute(self, task)
        return self.if_handler(task)


class Form(Node):
    def __init__(self, form, template_name=None, context=None, **kwargs):
        super().__init__(**kwargs)
        self._form = form
        self._context = context if context is not None else {}
        self._template_name = template_name
        self._user = None
        self._group = None
        self._permissions = []
        self._form_kwargs = {}
        self._run = True

    def execute(self, task: models.AutomationTaskModel):
        task = super().execute(task)

        if task is not None:
            if self._user is None and self._group and not self._permissions:
                raise ImproperlyConfigured("From: at least one .User, .Group, .Permission has to be specified")
            task.interaction_user = self.get_user()
            task.interaction_group = self.get_group()
            if task.data.get(f"_{self._name}_validated", None) is None:
                self.release_lock(task)  # Release lock and stop automation until form is validated
                return None
        return task  # Continue with validated form

    def validate(self, task: models.AutomationTaskModel, request, form):
        task.automation.data[f"_{self._name}_validated"] = request.user.id
        task.automation.save()

    def User(self, **kwargs):
        if self._user is not None:
            raise ImproperlyConfigured("Only one .User modifier for Form")
        self._user = kwargs
        return self

    def Group(self, **kwargs):
        if self._group is not None:
            raise ImproperlyConfigured("Only one .Group modifier for Form")
        self._group = kwargs
        return self

    def get_user(self):
        return User.objects.get(**self._user) if self._user is not None else None

    def get_group(self):
        return Group.objects.get(**self._group) if self._group is not None else None

    def Permission(self, permission):
        self._permissions.append(permission)
        return self

    def get_users_with_permission(self):
        perm = Permission.objects.filter(codename__in=self._permissions)
        filter = Q(groups__permissions__in=perm) | Q(user_permissions__in=perm)
        if self._user is not None:
            filter = filter & Q(**self._user)
        if self._group is not None:
            filter = filter & Q(group_set__contains=self._group)
        users = User.objects.filter(filter).distinct()
        return users


class ModelForm(Form):
    def __init__(self, form, key, template_name=None, **kwargs):
        super().__init__(form, template_name, **kwargs)
        if hasattr(Automation, key):
            raise ImproperlyConfigured(f"Chose different key for ModelForm node: {key} is a property of "
                                       f"flow.Automations")
        self._instance_key = key
        self._form_kwargs = self.get_model_from_kwargs

    def get_model_from_kwargs(self, task):
        model = self._form.Meta.model  # Get model from Form's Meta class
        if self._instance_key in task.data:
            instances = model.objects.filter(id=task.data[self._instance_key])
            return dict(instance=instances[0] if len(instances) == 1 else None)
        else:
            return dict()


class SendMessage(Node):
    def __init__(self, target, message, token=None, allow_multiple_receivers=False, **argv):
        self._target = target
        self._message = message
        self._token = token
        self._allow_multiple_receivers = allow_multiple_receivers
        self.argv = argv
        super().__init__()

    @on_execution_path
    def send_handler(self, task):
        cls = self._target
        if isinstance(cls, str):
            cls = models.get_automation_class(cls)
        if issubclass(cls, Automation):
            qs = models.AutomationModel.objects.filter(
                finished=False,
                automation_class=cls.__module__+"."+cls.__name__
            )
            automations = (cls(automation_id=instance.id) for instance in qs)
        elif isinstance(cls, Automation) or isinstance(cls, int):
            automations = (cls, )
        else:
            raise ImproperlyConfigured("")
        results = []
        for automation in automations:
            result = cls.dispatch_message(automation, self._message, self._token, data=self.argv)
            results.append((automation.id, result))
            if result is not None and not self._allow_multiple_receivers:
                break
        else:
            self.store_result(task, "No receiver", dict())
            return task
        self.store_result(task, "OK", dict(results=result))
        return task

    def execute(self, task: models.AutomationTaskModel):
        task = super().execute(task)
        return self.send_handler(task)

#
# Automation class
#
#


def on_signal(signal, start=None, **kwargs):
    """decorator for automations to connect to Django signals"""
    def decorator(cls):
        cls.on(signal, start, **kwargs)
        return cls
    return decorator


class Automation:
    model_class = models.AutomationModel
    singleton = False

    end = End()  # Endpoint of automation

    def __init__(self, **kwargs):
        super().__init__()
        prev = None
        self._iter = {}
        self._start = {}
        for name, attr in self.__class__.__dict__.items():
            if isinstance(attr, Node):  # Init nodes
                self._iter[prev] = attr
                attr.ready(self, name)
                prev = attr
            if isinstance(attr, ModelBase):  # Attach to Model instance
                at_c = copy(attr)  # Create copies of name and Model (attr)
                nm_c = copy(name)
                setattr(self.__class__, name,  # Replace property by get_model_instance
                        property(lambda slf: slf.get_model_instance(at_c, nm_c), self))
                if name not in kwargs:
                    kwargs[name] = None
                elif not isinstance(kwargs[name], int):  # Convert instance to id
                    kwargs[name] = kwargs[name].id
        self._iter[prev] = None  # Last item
        autorun = kwargs.pop("autorun", True)
        if 'automation' in kwargs:
            self._db = kwargs.pop('automation')
            assert isinstance(self._db, models.AutomationModel), \
                "automation= parameter needs to be AutomationModel instance"
            return
        elif 'automation_id' in kwargs:  # Attach to automation in DB
            self._db, _ = self.model_class.objects.get_or_create(
                id=kwargs.pop('automation_id'),
                defaults=dict(
                    automation_class=self.get_automation_class_name(),
                    finished=False,
                    data=kwargs,
                ))
            return
        elif self.singleton is True:  # Create or get singleton in DB
            self._db, _ = self.model_class.objects.get_or_create(
                automation_class=self.get_automation_class_name(),
            )
            self._db.data = kwargs
            self._db.finished = False
            self._db.save()
        elif self.singleton:
            assert isinstance(self.singleton, (list, tuple)), ".singleton can be bool, list, tuple or None"
            for key in self.singleton:
                assert key not in ("automation", "automation_id", "autorun"), \
                    f"'{key}' cannot be parameter to distinguish singleton automations. Chose a different name."
                assert key in kwargs, "to ensure singleton property, " \
                                      "create automation with '%s=...' parameter" % key
            qs = self.model_class.objects.filter(finished=False)
            for instance in qs:
                identical = sum((0 if key not in instance.data or instance.data[key] != kwargs[key] else 1
                                  for key in self.singleton))
                if identical == len(self.singleton):
                    self._db = instance
                    break
            else:
                self._db = self.model_class.objects.create(
                    automation_class=self.get_automation_class_name(),
                    finished=False,
                    data=kwargs,
                )
        else:
            self._db = self.model_class.objects.create(
                automation_class=self.get_automation_class_name(),
                finished=False,
                data=kwargs,
            )
        assert self._db is not None, "Internal error"
        if autorun:
            self.run()

    def get_model_instance(self, model, name):
        if not hasattr(self, '_' + name):
            setattr(self, '_' + name, model.objects.get(id=self._db.data[name]))
        return getattr(self, '_' + name)

    def get_automation_class_name(self):
        return self.__module__ + '.' + self.__class__.__name__

    @property
    def id(self):
        return self._db.id

    @property
    def data(self):
        assert self._db is not None, "Automation not bound to database"
        return self._db.data

    def save(self, *args, **kwargs):
        return self._db.save(*args, **kwargs)

    def nice(self, task=None, next_task=None):
        """Run automation steps in a background thread to, e.g., do not block
        the request response cycle"""
        threading.Thread(target=self.run,
                         kwargs=dict(task=task, next_task=next_task)
                         ).start()

    def run(self, task=None, next_node=None):
        """Execute automation until external responses are necessary"""
        assert not self.finished(), ValueError("Trying to run an already finished automation")

        if next_node is None:
            last_tasks = self._db.automationtaskmodel_set.filter(finished=None)
            if len(last_tasks) == 0:  # Start
                last, next_node = None, self._iter[None]  # First
            else:
                for last_task in last_tasks:
                    node = getattr(self, last_task.status)
                    self.run(last_task.previous, node)
                return

        while next_node is not None:
            task = next_node.enter(task)
            task = next_node.execute(task)
            last, next_node = task, next_node.leave(task)
        return last

    @classmethod
    def get_verbose_name(cls):
        if hasattr(cls, 'Meta'):
            if hasattr(cls.Meta, 'verbose_name'):
                return cls.Meta.verbose_name
        return f"Automation {cls.__name__}"

    @classmethod
    def get_verbose_name_plural(cls):
        if hasattr(cls, 'Meta'):
            if hasattr(cls.Meta, 'verbose_name_plural'):
                return cls.Meta.verbose_name_plural
        return f"Automations {cls.__name__}"

    def finished(self):
        return self._db.finished

    def start_from_request(self, request):              # pragma: no cover
        pass

    def send_message(self, message, token, data=None): # pragma: no cover
        """RECEIVES message and dispatches it within the class
        Called send_message so that sending a message to an automation
        is `automation.send_message(...)"""
        if hasattr(self, "receive_"+message):
            method = getattr(self, "receive_"+message)
            return method(token, data)
        return None

    @classmethod
    def satisfies_data_requirements(cls, message, get):
        if hasattr(cls, "receive_" + message):
            method = getattr(cls, "receive_" + message)
            if not hasattr(method, "data_requirements"):
                return False
            if hasattr(get, 'GET'):
                get = get.GET
            for param, type_class in method.data_requirements.items():
                if param not in get:
                    return False
                if not isinstance(get[param], type_class):  # Try simple conversion
                    try:
                        type_class(get[param])
                    except (ValueError, TypeError):
                        return False
            return True
        return False

    def kill(self):
        self._db.delete()

    @classmethod
    def dispatch_message(cls, automation, message, token, data):
        if isinstance(automation, int):
            automation = cls(automation_id=automation)
        assert isinstance(automation, cls), f"Wrong class to dispatch message: " \
                                            f"{automation.__class__.__name__} found, " \
                                            f"{cls.__name__} expected"
        return automation.send_message(message, token, data)

    @classmethod
    def create_on_message(cls, message, token, data):
        if cls.satisfies_data_requirements(message, data):
            kwargs = dict()
            accessor = data.GET if hasattr(data, 'GET') else data
            if isinstance(cls.singleton, (list, tuple)):
                for param in cls.singleton:
                    if param in accessor:
                        kwargs[param] = accessor.get(param)
            instance = cls(**kwargs)
            instance.send_message(message, token, data)
            return instance
        return None

    @classmethod
    def on(cls, signal, start=None, **kwargs):
        def creator(sender, **sargs):
            cls.on_signal(start, sender, **sargs)
        signal.connect(creator, weak=False, **kwargs)

    @classmethod
    def on_signal(cls, start, sender, **kwargs):
        instance = cls()  # Instantiate class
        if hasattr(instance, 'started_by_signal') and callable(instance.started_by_signal):
            instance.started_by_signal(sender, kwargs)  # initialize based on sender data
        instance.run(None, None if start is None else getattr(instance, start))  # run

    def __str__(self):
        return self.__class__.__name__


def get_automations(app=None):
    def check_module(mod):
        for item in dir(mod):
            attr = getattr(mod, item)
            if isinstance(attr, type) and issubclass(attr, Automation):
                automation_list.append((attr.__module__+'.'+ attr.__name__, attr.get_verbose_name()))

    automation_list = []
    if app is None:
        for name, mod in sys.modules.items():
            if name.rsplit(".")[-1] == "automations":
                check_module(mod)
    else:
        apps = app.split(".")

        mod = __import__(apps[0])
        for next in apps[1:]:
            mod = getattr(mod, next)

        check_module(mod)
        if hasattr(mod, 'automations'):
            mod = getattr(mod, 'automations')
            check_module(mod)

    return automation_list


def require_get_parameters(**kwargs):
    def decorator(method):
        method.data_requirements = kwargs
        return method
    return decorator



from django.apps import AppConfig
from django.utils.translation import gettext as _


class CmsAutomationsConfig(AppConfig):
    name = 'cms_automations'
    default_auto_field = 'django.db.models.AutoField'
    verbose_name = _('CMS Automations')



# coding=utf-8
from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class AutomationsConfig(AppConfig):
    name = 'automations'
    verbose_name = _('Automations')
    default_auto_field = 'django.db.models.AutoField'


# coding=utf-8

import sys
from logging import getLogger

from django import forms
from django.contrib.auth.models import Permission, User
from django.db.models import Q, QuerySet
from django.utils.module_loading import import_string
from django.utils.timezone import now
from django.utils.translation import gettext as _
from django.db import models

from . import settings

# Create your models here.

logger = getLogger(__name__)


def get_automation_class(dotted_name):
    components = dotted_name.split('.')
    cls = __import__(components[0])
    for path in components[1:]:
        cls = getattr(cls, path)
    return cls


class AutomationModel(models.Model):
    automation_class = models.CharField(
            max_length=256,
            blank=False,
            verbose_name=_("Process class"),
    )
    finished = models.BooleanField(
            default=False,
            verbose_name=_("Finished"),
    )
    data = models.JSONField(
            verbose_name=_("Data"),
            default=dict,
    )
    paused_until = models.DateTimeField(
            null=True,
            verbose_name=_("Paused until"),
    )
    created = models.DateTimeField(
            auto_now_add=True,
    )
    updated = models.DateTimeField(
            auto_now=True,
    )

    _automation_class = None

    def get_automation_class(self):
        if self._automation_class is None:
            self._automation_class = get_automation_class(self.automation_class)
        return self._automation_class

    @classmethod
    def run(cls, timestamp=None):
        if timestamp is None:
            timestamp = now()
        automations = cls.objects.filter(
                finished=False,
        ).filter(Q(paused_until__lte=timestamp) | Q(paused_until=None))

        for automation in automations:
            klass = import_string(automation.automation_class)
            instance = klass(automation_id=automation.id)
            logger.info(f"Running automation {automation.automation_class}")
            try:
                instance.run()
            except Exception as e:          # pragma: no cover
                automation.finished = True
                automation.save()
                logger.error(f'Error: {repr(e)}', exc_info=sys.exc_info())

    def __str__(self):
        return f"<AutomationModel for {self.automation_class}>"

class AutomationTaskModel(models.Model):
    automation = models.ForeignKey(
            AutomationModel,
            on_delete=models.CASCADE,
    )
    previous = models.ForeignKey(
            'automations.AutomationTaskModel',
            on_delete=models.SET_NULL,
            null=True,
            verbose_name=_("Previous task"),
    )
    status = models.CharField(
            max_length=256,
            blank=True,
            verbose_name=_("Status"),
    )
    locked = models.IntegerField(
            default=0,
            verbose_name=_("Locked"),
    )
    interaction_user = models.ForeignKey(
            'auth.User',
            null=True,
            on_delete=models.PROTECT,
            verbose_name=_("Assigned user"),
    )
    interaction_group = models.ForeignKey(
            'auth.Group',
            null=True,
            on_delete=models.PROTECT,
            verbose_name=_("Assigned group"),
    )
    interaction_permissions = models.JSONField(
        default=list,
        verbose_name=_("Required permissions"),
        help_text=_("List of permissions of the form app_label.codename"),
    )
    created = models.DateTimeField(
            auto_now_add=True,
    )
    finished = models.DateTimeField(
            null=True,
    )
    message = models.CharField(
            max_length=settings.MAX_FIELD_LENGTH,
            verbose_name=_("Message"),
            blank=True,
    )
    result = models.JSONField(
            verbose_name=_("Result"),
            null=True,
            blank=True,
            default=dict,
    )

    @property
    def data(self):
        return self.automation.data

    def get_users_with_permission(self, include_superusers=True,
                                  backend="django.contrib.auth.backends.ModelBackend"):

        users = User.objects.all()
        for permission in self.interaction_permissions:
            users &= User.objects.with_perm(permission, include_superusers=False, backend=backend)
        if self.interaction_user is not None:
            users = users.filter(id=self.interaction_user_id)
        if self.interaction_group is not None:
            users = users.filter(groups=self.interaction_group)
        if include_superusers:
            users |= User.objects.filter(is_superuser=True)
        return users

    def get_node(self):
        cls = self.automation.get_automation_class()
        instance = cls(automation=self.automation)
        return getattr(instance, self.status)

    @classmethod
    def get_open_tasks(cls, user):
        candidates = cls.objects.filter(finished=None)
        return tuple(task for task in candidates if user in task.get_users_with_permission())




from logging import getLogger

from django.core.management import BaseCommand

from automations.models import AutomationModel


logger = getLogger(__name__)


class Command(BaseCommand):
    help = 'Touch every automation to proceed.'

    def handle(self, *args, **options):
        AutomationModel.run()